# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from abc import ABCMeta, abstractmethod

import six


@six.add_metaclass(ABCMeta)
class ConfigManager(object):
    """
    Abstract class for all config managers.
    Use this class to define custom configuration pulling logic.
    For example, if you need to merge all configurations in the same folder.
    """

    @abstractmethod
    def iterconfigs(self):
        """Similar to iteritems method. Iterate through all configurations, this will not cache results."""
        pass

    @abstractmethod
    def resolve(self, path, **filters):
        pass



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from garlicconfig.exceptions import ValidationError
from garlicconfig.models import ConfigModel

from . import ConfigField, assert_value_type


class ModelField(ConfigField):

    def __init__(self, model_class, **kwargs):
        """
        A field that stores another config field as a subsection.
        :param model_class: Any class of type ConfigModel to store as a subsection.
        :type model_class: Type[ConfigModel]
        """
        if not isinstance(model_class, type):
            raise TypeError("'model_class' has to be a type.")
        if not issubclass(model_class, ConfigModel):
            raise ValueError("'model_class' has to implement ConfigModel")
        self.model_class = model_class
        instance = self.model_class()  # initialize an instance
        kwargs['default'] = instance
        super(ModelField, self).__init__(**kwargs)

    def validate(self, value):
        super(ModelField, self).validate(value)
        assert_value_type(value, self.model_class, self.name)

    def to_dict_value(self, value):
        dict_data = value.get_dict() if value else None
        return dict_data if dict_data else None  # if data is empty, return None

    def to_model_value(self, value):
        if not value:
            return self.model_class()  # initialize a new instance
        if not isinstance(value, dict):
            raise ValidationError("Value for {key} must be a python dict.".format(key=self.name))
        return self.model_class.load_dict(value)

    def __extra_desc__(self):
        return {
            'model_info': {
                'name': self.model_class.__name__,
                'fields': self.model_class.get_model_desc_dict(),
            }
        }



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from garlicconfig.exceptions import ValidationError
from garlicconfig.utils import assert_value_type

import six


class ConfigField(object):
    """
    Abstract class for all config fields.
    """

    __init_args = ('name', 'nullable', 'default', 'desc',)

    def __init__(self, **kwargs):
        """
        Base configuration model, holds most basic data about a field.

        Parameters:
            default : the default value, note that it still has to pass all the validation tests.
            nullable (bool) : determines whether this field is permitted to have to no value at all.
            desc (str) : a short description of what this field is.
        """
        self.friendly_name = kwargs.get('name')
        self.nullable = kwargs.get('nullable', True)
        self.default = kwargs.get('default')
        self.desc = kwargs.get('desc')
        self.__validate__(self.default)  # make sure default value is valid

        # it's nice to raise an exception when we get an unexpected argument so it's extra clear we're not handling it.
        unrecognized_args = [arg for arg in kwargs if arg not in self.__init_args]
        if unrecognized_args:
            raise TypeError("Argument '{arg_name}' is not recognized.".format(arg_name=unrecognized_args[0]))

    @property
    def name(self):
        """
        :return: str for the field's name.
        """
        return self.friendly_name or type(self).__name__

    def __validate__(self, value):
        """
        Low level method for validation. Do not use this method outside of this package, it's prone to change without
        further notice.
        """
        if value is None and not self.nullable:
            raise ValidationError("Value for '{key}' is not allowed to be null.".format(key=self.name))
        elif value is not None:
            self.validate(value)

    def validate(self, value):
        """
        Determines whether or not the given value is valid for the current field.

        Parameters:
            value : value is guaranteed to be non-null.
        """
        pass

    def to_model_value(self, value):
        """
        Given a value from a python dictionary, return the appropriate value to store in the model
        For example, if you have a custom class, you should use this method to initialize it.
        """
        return value

    def to_dict_value(self, value):
        """
        Given a model value, return a basic value type to be stored in a python dictionary.
        Note that this value must only hold basic types: list of integers is acceptable but a custom class is not.
        """
        return value

    def get_field_desc_dict(self):
        """
        Generate a dictionary describing the structure of this field.
        """
        extra_obj = self.__extra_desc__() or {}
        extra_obj.update({
            'nullable': self.nullable,
            'default': self.to_dict_value(self.default),
        })
        return {
            'name': self.name,
            'type': type(self).__name__,
            'desc': self.desc,
            'extra': extra_obj,
        }

    def __extra_desc__(self):
        """
        If there is any extra object this field needs to have, return it here.
        It'll be included in the 'extra' field returned in the field description dictionary.
        """
        pass


class StringField(ConfigField):

    def __init__(self, choices=None, **kwargs):
        """
        :param choices: If provided, value has to be one of these values.
        :type choices: iterator
        """
        if choices and not hasattr(choices, '__iter__'):
            raise TypeError("'choices' has to be a sequence of string elements.")
        self.choices = choices
        super(StringField, self).__init__(**kwargs)

    def validate(self, value):
        super(StringField, self).validate(value)
        assert_value_type(value, six.text_type, self.name)
        if self.choices and value not in self.choices:
            raise ValidationError("Value '{given}' for '{key}' is not accepted. Choices are '{choices}'".format(
                given=value,
                key=self.name,
                choices="', '".join(self.choices)
            ))

    def __extra_desc__(self):
        if self.choices:
            return {
                'choices': self.choices,
            }


class BooleanField(ConfigField):

    def validate(self, value):
        super(BooleanField, self).validate(value)
        assert_value_type(value, bool, self.name)


class IntegerField(ConfigField):

    def __init__(self, domain=None, **kwargs):
        """
        :param domain: Specify the domain of accepted values as (min, max,)
        :type domain: tuple
        """
        if domain and (not isinstance(domain, tuple) or len(domain) != 2):
            raise TypeError("'domain' has to be a tuple providing inclusive domain like: (min,max)")
        self.domain = domain
        super(IntegerField, self).__init__(**kwargs)

    def validate(self, value):
        super(IntegerField, self).validate(value)
        assert_value_type(value, int, self.name)
        if self.domain and not (self.domain[0] <= value <= self.domain[1]):
            raise ValidationError(
                "Value '{value}' for '{key}' has to be in range {domain}.".format(
                    value=value,
                    key=self.name,
                    domain=self.domain
                )
            )

    def __extra_desc__(self):
        if self.domain:
            return {
                'domain': self.domain,
            }


class ArrayField(ConfigField):

    def __init__(self, field, **kwargs):
        """
        Stores an array of ConfigField(s).
        :param field: An ConfigField instance describing the values in the list.
        :type field: ConfigField
        """
        if not isinstance(field, ConfigField):
            raise TypeError("'field' has to be a ConfigField.")
        self.field = field
        super(ArrayField, self).__init__(**kwargs)

    def validate(self, value):
        super(ArrayField, self).validate(value)
        assert_value_type(value, list, self.name)
        for item in value:
            self.field.validate(item)

    def to_model_value(self, value):
        return list(map(lambda x: self.field.to_model_value(x), value)) if value else None

    def to_dict_value(self, value):
        return list(map(lambda x: self.field.to_dict_value(x), value)) if value else None

    def __extra_desc__(self):
        return {
            'element_info': self.field.get_field_desc_dict()
        }



# -*- coding: utf-8 -*-
from __future__ import unicode_literals


class ValidationError(Exception):
    """The provided value is not valid."""
    pass


class ConfigNotFound(Exception):
    """No configuration with such key name was found."""
    pass



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import json
from abc import ABCMeta, abstractmethod

from garlicconfig.models import ConfigModel

import six


@six.add_metaclass(ABCMeta)
class ConfigEncoder(object):

    @abstractmethod
    def encode(self, config, pretty=True):
        pass

    @abstractmethod
    def decode(self, data, config_class):
        pass


class PrettyJSONEncoder(json.JSONEncoder):
    """
    This encoder is not the most efficient at all, but it produces a pretty json format. It solves the problem with
    long arrays. Default encoder puts every single element on a separate line but this encoder does a smart job.
    """
    def __init__(self, *args, **kwargs):
        super(PrettyJSONEncoder, self).__init__(*args, **kwargs)
        self.current_indent = 0
        self.current_indent_str = ''

    @staticmethod
    def __has_non_primitive(obj):
        # since this is a private func, it's safe to assume obj is an iterable.
        for item in obj:
            if isinstance(item, (list, tuple, dict)):
                return True
        return False

    def encode(self, obj):
        if isinstance(obj, (list, tuple)):
            primitives_only = not self.__has_non_primitive(obj)
            output = []
            if primitives_only:
                for item in obj:
                    output.append(super(PrettyJSONEncoder, self).encode(item))
                return '[' + ', '.join(output) + ']'
            else:
                self.current_indent += self.indent
                self.current_indent_str = ''.join([' ' for x in range(self.current_indent)])
                for item in obj:
                    output.append(self.current_indent_str + self.encode(item))
                self.current_indent -= self.indent
                self.current_indent_str = ''.join([' ' for x in range(self.current_indent)])
                return '[\n' + ',\n'.join(output) + '\n' + self.current_indent_str + ']'
        elif isinstance(obj, dict):
            output = []
            self.current_indent += self.indent
            self.current_indent_str = ''.join([' ' for x in range(self.current_indent)])
            for key in sorted(obj):
                value = obj[key]
                output.append(
                    self.current_indent_str + super(PrettyJSONEncoder, self).encode(key) + ': ' + self.encode(value)
                )
            self.current_indent -= self.indent
            self.current_indent_str = "".join([" " for x in range(self.current_indent)])
            return '{\n' + ',\n'.join(output) + '\n' + self.current_indent_str + '}'
        else:
            return super(PrettyJSONEncoder, self).encode(obj)


class JsonEncoder(ConfigEncoder):

    def encode(self, config, pretty=True):
        if not isinstance(config, ConfigModel):
            raise TypeError("'config' must be a ConfigModel.")
        if pretty:
            return json.dumps(config.get_dict(), sort_keys=True, indent=2, cls=PrettyJSONEncoder)
        else:
            return json.dumps(config.get_dict(), sort_keys=True, separators=(',', ':'))

    def decode(self, data, config_class):
        if not issubclass(config_class, ConfigModel):
            raise TypeError("'config_class' must be a ConfigModel.")
        data_dict = json.loads(data)
        return config_class.load_dict(data_dict)


def encode(config, cls=None, pretty=True):
    """
    Encodes a config instance.
    :param config: Any instance of ConfigModel.
    :type config: ConfigModel
    :param cls: The encoder class. Must be of type ConfigEncoder.
    :type cls: ConfigEncoder
    :param pretty: Whether or not the encoder should attempt to return a human readable format.
    """
    cls = cls or JsonEncoder  # default to json
    if not issubclass(cls, ConfigEncoder):
        raise TypeError("'cls' must be a ConfigEncoder")
    return cls().encode(config, pretty)


def decode(data, config_class, cls=None):
    """
    Decodes the raw data using a decoder to a config instance.
    :param data: The raw data to be processed by the encoder.
    :param config_class: The config model class, to be used as a model for validation.
    :type config_class: ConfigModel
    :param cls: The encoder class. Must be of type ConfigEncoder.
    """
    cls = cls or JsonEncoder
    if not issubclass(cls, ConfigEncoder):
        raise TypeError("'cls' must be a ConfigEncoder")
    return cls().decode(data, config_class)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import os
import re
from abc import ABCMeta, abstractmethod

from garlicconfig.exceptions import ConfigNotFound

import six


@six.add_metaclass(ABCMeta)
class ConfigRepository(object):

    @abstractmethod
    def all(self):
        """
        Returns a generator containing the name of all available configs.
        """
        pass

    @abstractmethod
    def retrieve(self, name):
        """
        Returns the string data for this config if it exists.
        """
        pass

    @abstractmethod
    def save(self, name, data):
        """
        Saves the string data for the given config.
        """
        pass


class FileConfigRepository(ConfigRepository):

    def __init__(self, root_dir):
        self.root = root_dir

    def all(self):
        rule = re.compile('(.*)\.garlic')
        for fname in os.listdir(self.root):
            config_name = rule.match(fname)
            if not config_name:
                continue
            yield config_name.group(1)

    def retrieve(self, name):
        try:
            with open(os.path.join(self.root, name + '.garlic'), 'r') as config_data:
                return config_data.read()
        except IOError:
            raise ConfigNotFound()

    def save(self, name, data):
        with open(os.path.join(self.root, name + '.garlic'), 'w') as config_file:
            config_file.write(data)


class MemoryConfigRepository(ConfigRepository):

    def __init__(self):
        self.storage = {}

    def all(self):
        return self.storage.keys()

    def retrieve(self, name):
        if name in self.storage:
            return self.storage[name]
        else:
            raise ConfigNotFound

    def save(self, name, data):
        self.storage[name] = data



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import copy

from garlicconfig.exceptions import ValidationError


def assert_value_type(value, expected_type, name):
    if not isinstance(value, expected_type):
        raise ValidationError(
            "Expected '{expected}' for '{key}', but got '{got}'.".format(
                expected=expected_type.__name__,
                key=name,
                got=type(value).__name__
            )
        )


def __merge(base, config):
    for key in config:
        if isinstance(config[key], dict) and key in base:
            __merge(base[key], config[key])
        else:
            base[key] = config[key]
    return base


def merge(base, config):
    """
    Merge two configurations.

    Parameters:
        base: dict representation of the configuration.
        config: the configuration that will override base config.
    """
    base = base or {}
    config = config or {}
    assert_value_type(base, dict, 'base')
    assert_value_type(config, dict, 'config')
    return __merge(copy.deepcopy(base), config)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import copy

from garlicconfig.exceptions import ValidationError
from garlicconfig.fields import ConfigField

import six


class ModelMetaInfo(object):

    def __init__(self):
        self.fields = {}  # map name to fields


class ModelMetaClass(type):

    def __new__(mcs, name, bases, attributes):
        new_class = super(ModelMetaClass, mcs).__new__(mcs, str(name), bases, attributes)
        meta = ModelMetaInfo()
        for key in attributes:
            field = attributes[key]
            if isinstance(attributes[key], ConfigField):
                # if a friendly name is provided, skip this step.
                if not field.friendly_name:
                    field.friendly_name = key
                meta.fields[key] = field
                delattr(new_class, key)
        for base in bases:
            if isinstance(base, ModelMetaClass):
                meta.fields.update(base.__meta__.fields)
        setattr(new_class, '__meta__', meta)
        return new_class


@six.add_metaclass(ModelMetaClass)
class ConfigModel(object):

    def __init__(self):
        for field_name in self.__meta__.fields:
            setattr(self, field_name, copy.deepcopy(self.__meta__.fields[field_name].default))

    @classmethod
    def load_dict(cls, obj):
        """
        Instantiate a config model and load it using the given dictionary.
        """
        new_instance = cls()
        for field_name in cls.__meta__.fields:
            field = cls.__meta__.fields[field_name]
            if obj and field_name in obj:
                value = field.to_model_value(obj[field_name])
                setattr(new_instance, field_name, value)
            elif not field.nullable:
                raise ValidationError("Value for '{key}' cannot be null.".format(key=field_name))
        return new_instance

    @classmethod
    def get_model_desc_dict(cls):
        """
        Returns a python dictionary containing description for the current model and its children.
        """
        obj = {}
        for field_name in cls.__meta__.fields:
            field = cls.__meta__.fields[field_name]
            obj[field_name] = field.get_field_desc_dict()
        return obj

    def get_dict(self):
        """
        Returns a python dictionary containing only basic types so it can be used for encoding.
        """
        obj = {}
        for field_name in self.__meta__.fields:
            model_value = getattr(self, field_name)
            dict_value = self.__meta__.fields[field_name].to_dict_value(model_value)
            if dict_value is None:
                continue
            obj[field_name] = dict_value
        return obj

    def __setattr__(self, name, value):
        if name in self.__meta__.fields:
            self.__meta__.fields[name].__validate__(value)
        super(ConfigModel, self).__setattr__(name, value)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from setuptools import find_packages, setup

with open('VERSION', 'r') as reader:
    version = reader.read().strip()


def get_long_description():
    with open('README.md', 'r') as reader:
        return reader.read()


setup(
    author='Peyman Mortazavi',
    author_email='pey.mortazavi@gmail.com',
    name='garlicconfig',
    packages=find_packages(),
    include_package_data=True,
    description='InfoScout GarlicConfig',
    long_description=get_long_description(),
    long_description_content_type='text/markdown; charset=UTF-8',
    url='https://github.com/infoscout/garlicconfig',
    download_url='https://github.com/infoscout/garlicconfig/archive/{version}.tar.gz'.format(version=version),
    version=version,
    install_requires=['six'],
    keywords=['configs', 'settings']
)
